#!/usr/bin/env python

import os, sys, pip, json, datetime, ntpath

__author__ = 'Hieu Nguyen'
__version__ = '1.0'

# ***********************************************  CONSTANTS  *********************************************** #

RESIDUAL_FILE_EXTENSION = '.res'
JSON_FILE_EXTENSION = '.json'

HDR_GRP = 'HDR GRP CANADIAN GEODETIC SURVEY, SURVEYOR GENERAL BRANCH, NATURAL RESOURCES CANADA'
HDR_ARP = 'HDR ADR GOVERNMENT OF CANADA, 588 BOOTH STREET ROOM 334, OTTAWA ONTARIO K1A 0Y7'
HDR_TEL = 'HDR TEL 343-292-6617'
HDR_EMA = 'HDR EMA nrcan.geodeticinformationservices.rncan@canada.ca'

HDR_LINES = [HDR_GRP, HDR_ARP, HDR_TEL, HDR_EMA]

WIDTH = {'DIR': 3, 'STN': 4, 'DAYofYEAR': 11, 'YEAR-MM-DD': 10, 'HH:MM:SS.SS': 11,
         'PRN': 3, 'SIG': 3, 'AZI': 6, 'ELV': 4, 'RES': 8, 'AMB': 14, 'SAM': 8, 'INI': 8}

RES_HEADER = ['DIR', 'STN', 'DAYofYEAR', 'YEAR-MM-DD', 'HH:MM:SS.SS',
              'PRN', 'SIG', 'AZI', 'ELV', 'RES', 'AMB', 'SAM', 'INI']

R_DIR = 'DIR'
R_STN = 'STN'
R_DAYofYEAR = 'DAYofYEAR'
R_YMD = 'YEAR-MM-DD'
R_HMS = 'HH:MM:SS.SS'
R_PRN = 'PRN'
R_SIG = 'SIG'
R_AZI = 'AZI'
R_ELV = 'ELV'
R_RES = 'RES'
R_AMB = 'AMB'
R_SAM = 'SAM'
R_INI = 'INI'

J_DIR = 'dir'
J_DOY = 'doy'
J_EPOCH = 'epoch'
J_SAT = 'sat'
J_AZ = 'az'
J_EL = 'el'
J_PRN = 'prn'
J_SIG = 'sig'
J_RES = 'res'
J_TYPE = 'type'
J_AMB = 'amb'
J_INI = 'ini'
J_SAMB = 'samb'
J_STA = 'sta'

# ***********************************************  CLASS DEF  *********************************************** #


class Json2Res:

    def __init__(self, filename, destination):
        self.filename, self.file_extension = os.path.splitext(filename)
        self.destination = destination
        self.json_data = None

        self.load()

    def load(self):
        resfilename = os.path.join(self.destination, ntpath.basename(self.filename) + RESIDUAL_FILE_EXTENSION)

        try:
            with open(self.filename + self.file_extension) as f:
                # load json array
                self.json_data = json.loads(f.read())

            with open(resfilename, 'w+') as f:
                for hdr_line in HDR_LINES:
                    f.write(hdr_line + '\n')

                # write column headers
                f.write(self.get_res_header() + '\n')

                # iterate json objects in array
                for json_obj in self.json_data:
                    for signal_line in self.format_object(json_obj):
                        f.write(signal_line + '\n')

            return {'status': 'pass', 'file': resfilename, 'exception': None}
        except Exception as e:
            return {'status': 'fail', 'file': resfilename, 'exception': e}

    def format_object(self, json_obj):
        signal_list = []

        _dir = ('{:>' + str(WIDTH[R_DIR]) + '}').format(json_obj[J_DIR])
        _stn = ('{:>' + str(WIDTH[R_STN]) + '}').format(json_obj[J_STA])
        _doy = ('{:>' + str(WIDTH[R_DAYofYEAR]) + '.7f}').format(float(json_obj[J_DOY]))

        _epoch = json_obj[J_EPOCH].split()
        _ymd = ('{:>' + str(WIDTH[R_YMD]) + '}').format(_epoch[0])
        _hms = ('{:>' + str(WIDTH[R_HMS]) + '}').format(_epoch[1])

        for satellite in json_obj[J_SAT]:
            _prn = ('{:>' + str(WIDTH[R_PRN]) + '}').format(satellite[J_PRN])

            _azi = ('{:>' + str(WIDTH[R_AZI]) + '}').format(satellite[J_AZ])
            _elv = ('{:>' + str(WIDTH[R_ELV]) + '}').format(satellite[J_EL])

            for signal in satellite[J_SIG]:
                _sig = ('{:>' + str(WIDTH[R_SIG]) + '}').format(signal[J_TYPE])
                _res = ('{:>' + str(WIDTH[R_RES]) + '}').format(signal[J_RES])

                _amb = ('{:>' + str(WIDTH[R_AMB]) + '.3f}').format(
                    float(signal[J_AMB]) if J_AMB in signal else 0.0
                )

                _sam = ('{:>' + str(WIDTH[R_SAM]) + '.3f}').format(
                    float(signal[J_SAMB]) if J_SAMB in signal else 0.0
                )
                _ini = ('{:>' + str(WIDTH[R_INI]) + '}').format(
                    int(signal[J_INI]) if J_INI in signal else 0
                )

                signal_list.append(' '.join([_dir, _stn, _doy, _ymd, _hms,
                                             _prn, _sig, _azi, _elv, _res, _amb, _sam, _ini]))

        return signal_list

    def get_res_header(self):
        return ' '.join([('{:>' + str(WIDTH[k]) + '}').format(k) for k in RES_HEADER])


class Res2Json:

    def __init__(self, filename, destination):
        self.filename, self.extension = os.path.splitext(filename)
        self.destination = destination

        self.load()

    def load(self):
        json_array = []
        resfilename = os.path.join(self.destination, ntpath.basename(self.filename) + JSON_FILE_EXTENSION)

        try:
            with open(self.filename + self.extension, 'r') as f:

                current_epoch = None
                json_object = {}

                for line in f:
                    # skip header lines
                    if line.startswith('HDR') or line.startswith('DIR'):
                        continue

                    line_data = dict(zip(RES_HEADER, line.split()))
                    line_epoch = ' '.join([line_data[R_YMD], line_data[R_HMS]])

                    if current_epoch is None:
                        # load first line
                        json_object = self.parse_new_epoch(line_data)
                        current_epoch = line_epoch
                    else:

                        # new epoch
                        if not current_epoch == line_epoch:
                            # add current json object to array
                            json_array.append(json_object)

                            # load new json object from new epoch line
                            json_object = self.parse_new_epoch(line_data)

                            current_epoch = line_epoch
                        else:
                            # current line has the same epoch as current json object
                            self.parse_satellite(json_object, line_data)

                # add last json object
                json_array.append(json_object)

            with open(resfilename, 'w+') as f:
                json.dump(json_array, f)

            return {'status': 'pass', 'file': resfilename, 'exception': None}
        except Exception as e:
            return {'status': 'fail', 'file': resfilename, 'exception': e}

    def parse_new_epoch(self, kwargs):
        signal = {
            J_RES: kwargs[R_RES],
            J_TYPE: kwargs[R_SIG]
        }

        if float(kwargs[R_AMB]) != 0.0 or float(kwargs[R_SAM]) != 0.0 or int(kwargs[R_INI]) != 0:
            signal[J_AMB] = kwargs[R_AMB]
            signal[J_SAMB] = kwargs[R_SAM]
            signal[J_INI] = kwargs[R_INI]

        satellite = {
            J_AZ: kwargs[R_AZI],
            J_EL: kwargs[R_ELV],
            J_PRN: kwargs[R_PRN],
            J_SIG: [signal]
        }
        return {J_DIR: kwargs[R_DIR], J_DOY: kwargs[R_DAYofYEAR],
                J_EPOCH: ' '.join([kwargs[R_YMD], kwargs[R_HMS]]),
                J_SAT: [satellite],
                J_STA: kwargs[R_STN]}

    def parse_satellite(self, json_obj, kwargs):
        satellite = None

        for sat in json_obj[J_SAT]:
            if sat[J_PRN] == kwargs[R_PRN]:
                satellite = sat

        signal = {
            J_RES: kwargs[R_RES],
            J_TYPE: kwargs[R_SIG]
        }

        if float(kwargs[R_AMB]) != 0.0 or float(kwargs[R_SAM]) != 0.0 or int(kwargs[R_INI]) != 0:
            signal[J_AMB] = kwargs[R_AMB]
            signal[J_SAMB] = kwargs[R_SAM]
            signal[J_INI] = kwargs[R_INI]

        if satellite is None:
            # add new satellite
            json_obj[J_SAT].append({
                J_AZ: kwargs[R_AZI],
                J_EL: kwargs[R_ELV],
                J_PRN: kwargs[R_PRN],
                J_SIG: [signal]
            })
        else:
            # add new signal to satellite
            satellite[J_SIG].append(signal)


# **********************************************  MAIN CLASS  ********************************************** #

def execute(_path_type, _path, _conversion_type, _output_path, v=False):
    if _path_type == 'f':
        if not os.path.isfile(_path):
            print("Path is not a file: " + _path)
            sys.exit(0)

        if not validate_file_type(_path, _conversion_type):
            sys.exit()

        if _conversion_type == 'j':
            Res2Json(_path, _output_path)
        elif _conversion_type == 'r':
            Json2Res(_path, _output_path)
        else:
            if get_file_type(_path) == 'r':
                Res2Json(_path, _output_path)
            elif get_file_type(_path) == 'j':
                Json2Res(_path, _output_path)
            else:
                print("Error: incorrect file format.")
                sys.exit()

    else:
        if not os.path.isdir(_path):
            print("Path is not a directory: " + _path)
            sys.exit(0)

        # Convert all files in directory
        for _file in os.listdir(_path):
            filename, ext = os.path.splitext(_file)
            _file = os.path.join(_path, _file)
            if _conversion_type == 'r':
                if ext == RESIDUAL_FILE_EXTENSION:
                    Json2Res(_file, _output_path)
                    if v:
                        print('+:' + _file)
                else:
                    if v:
                        print('-:' + _file + ' : not a json file.')
            elif _conversion_type == 'j':
                if ext == JSON_FILE_EXTENSION:
                    Res2Json(_file, _output_path)
                    if v:
                        print('+:' + _file)
                else:
                    if v:
                        print('-:' + _file + ' : not a res file.')
            else:
                if ext in [RESIDUAL_FILE_EXTENSION, JSON_FILE_EXTENSION]:
                    if get_file_type(_file) == 'r':
                        Res2Json(_file, _output_path)
                        if v:
                            print('+:' + _file)
                    elif get_file_type(_file) == 'j':
                        Json2Res(_file, _output_path)
                        if v:
                            print('+:' + _file)
                else:
                    if v:
                        print(_file + ' is not a valid res/json file.')


def handle_conversion_err(msg):
    if msg['status'] == 'fail':
        # type(msg['exception']).__name__ == KeyError
        print(msg['file'] + " failed.")


def confirm_overwrite():
    confirm = input('Do you wish to overwrite original files? (y/n)\n')
    return confirm.lower() in ['y', 'yes']


def confirm_continue():
    confirm = raw_input('This may results in overwriting original files.\nDo you wish to continue? (y/n)\n')
    return confirm.lower() in ['y', 'yes']


def get_file_type(filename):
    c = None
    with open(filename) as f:
        c = f.read(1)

    if c in ['{', '[']:
        return 'j'
    elif c in ['H', 'D']:
        return 'r'
    else:
        return None


def validate(_in_path, _path_type, _conversion_type, _output_path):
    if os.path.exists(_output_path):
        if _output_path == "ConvertedResiduals":
            _output_path += str(datetime.datetime.now().strftime('%Y%m%d%H%M%S'))
        if os.listdir(_output_path):
            print("Directory is not empty.")
            if not confirm_continue():
                sys.exit()
    else:
        os.mkdir(_output_path)


def validate_file_type(_in_path, _conversion_type):
    filename, ext = os.path.splitext(_in_path)
    abr = {'j': '.json', 'r': '.res', 'm': 'm'}
    if ext == abr[conversion_type]:
        detected_file_type = get_file_type(_in_path)
        if detected_file_type is None:
            print("Error: " + _in_path + " appears to be empty.")
            return 0
        if _conversion_type == detected_file_type:
            print(_in_path + " is already in " + abr[_conversion_type] + " format.")
            return 0
    return 1


def install(module):
    pip.main(['install', module])


if __name__ == '__main__':

    try:
        import argparse
    except ImportError:
        pip.main(['install', 'argparse'])
        import argparse

    # command line argument parser
    parser = argparse.ArgumentParser(description='Residual file converter. Transforms to and from, '
                                                 'ASCII (.res) format and JSON (.json) format. '
                                                 'ie "c:\> ResConverter -f residualfile.res -j -o ConvertedResiduals"')

    parser.add_argument('-v', '--verbose', action='store_true', help='Print verbose')

    pathType = parser.add_mutually_exclusive_group(required=True)
    pathType.add_argument('-f', '--file', help='File path')
    pathType.add_argument('-d', '--directory', help='Directory path (convert all valid file formats to counterpart)')

    conversionType = parser.add_mutually_exclusive_group()
    conversionType.add_argument('-r', '--res', action='store_true', help='Convert to ASCII residual format')
    conversionType.add_argument('-j', '--json', action='store_true', help='Convert to JSON file format')
    conversionType.add_argument('-a', '--auto', action='store_true', help='Try to dynamically convert file(s) (default)')

    parser.add_argument('-o', '--outputdir', help='Output directory (default "./ConvertedResiduals" will be created, if option is not given)')

    if not len(sys.argv) > 1:
        x = raw_input('Run from console. \nie. "c:\> ResConverter.exe -f residualfile.res -j -o ConvertedResiduals" \nPress Enter to exit.')
        sys.exit(0)

    parser.parse_args()

    args = parser.parse_args()

    path = args.file if args.directory is None else args.directory
    path_type = "f" if args.directory is None else "d"
    conversion_type = "j" if args.json else ("r" if args.res else "m")
    output_path = args.outputdir if args.outputdir is not None else "ConvertedResiduals"

    validate(path, path_type, conversion_type, output_path)

    execute(path_type, path, conversion_type, output_path, args.verbose)
